/*
 ķ
   Program..: _DCARRAY.PRG                                         
   Author...: Roger J. Donnay                                      
   Notice...: (c) DONNAY Software Designs 1987-2001                
   Date.....: Mar 24, 2001                                         
   Notes....: Functions for writing, reading and converting arrays 
                                                                   
   Functions: dc_array_r(), dc_array_w(), dc_aconvert()            
              dc_ascanrev(), dc_areverse(), dc_array2prg()         
              dc_abigelem(), dc_acompare(), dc_astrtran()          
              dc_arrayparse(), dc_arrayview()                      
 Ľ
*/

MEMVAR oObject

#include 'set.ch'
#include 'class.ch'
#include 'appevent.ch'
#include 'dcprint.ch'

#define DC_ARRMSG_0   'Invalid format in array file:'
#define DC_ARRMSG_1   ' Writing array to file: '
#define DC_ARRMSG_2   ' Please wait...'
#define DC_ARRMSG_3_1 ' File exists '
#define DC_ARRMSG_3_2 'Append'
#define DC_ARRMSG_3_3 'Overwrite'
#define DC_ARRMSG_3_4 'a~Bort'

#INCLUDE "dctree.ch"
#INCLUDE "dcicon.ch"
#INCLUDE "dcbitmap.ch"

MEMVAR aOutArray
STATIC cArrayFile := ''

* Syntax DC_ARRAY_R(<expC1>)
*
*  expC1 - Name of file
*  Returns an array restored from file <expC1>

* -------------------

FUNCTION dc_array_r ( cFileName )

LOCAL  cTextLine, cType, cElement, nHandle, nComma, cValue,;
       cArray, cPath, bErrorBlock
PRIVATE aOutArray := {}

DC_VALTYPE( @cFileName,'' )
cPath := DC_PATH( cFileName )
IF EMPTY( cPath )
  cPath := SET(_SET_DEFAULT)
ELSE
  cPath := ''
ENDIF
nHandle := DC_TXTOPEN( cPath + cFileName )
IF nHandle<=0
  RETURN nil
ENDIF
cValue := ''
cArrayFile := cPath + cFileName
bErrorBlock := ERRORBLOCK( {|e| _DCARRAYERROR(e)} )
BEGIN SEQUENCE
DO WHILE !DC_TXTEOF(nHandle)
  cTextLine := DC_TXTLINE(nHandle)
  DC_TXTSKIP(nHandle,1)
  IF LEFT(cTextLine,1) = '[' .OR. LEFT(cTextLine,1) = ','
    nComma := AT(',',cTextLine)
    cElement := TRIM(SUBSTR(cTextLine,1,nComma-1))
    cType := UPPER(SUBSTR(cTextLine,nComma+1,1))
    cValue := SUBSTR(cTextLine,nComma+3)
    cArray := 'aOutArray'+cElement
  ELSE
    cValue += cTextLine
  ENDIF
  DO CASE
    CASE cType='C'
      &cArray := STRTRAN(cValue,CHR(255),CHR(13)+CHR(10))
    CASE cType='N'
      &cArray := VAL(cValue)
    CASE cType='D'
      &cArray := CTOD(cValue)
    CASE cType='L'
      &cArray := IIF(cValue='Y',.t.,.f.)
    CASE cType='A'
      cArray := 'aOutArray'+cElement
      &cArray := {}
      ASIZE(&cArray,VAL(cValue))
    CASE cType='B'
      &cArray := &('"'+cValue+'"')
  ENDCASE
ENDDO
END SEQUENCE
ERRORBLOCK(bErrorBlock)
DC_TXTCLOSE(nHandle)
RETURN aOutArray

* -------------------

STATIC FUNC _dcarrayerror()

DC_MSGBOX(,,{DC_ARRMSG_0,cArrayFile},,.t.)
aOutArray := nil
BREAK
RETURN nil

* -------------------

FUNCTION dc_array_w ( aArray, cFileName )

LOCAL  nHandle, cPath

DC_VALTYPE( @cFileName,'' )
cPath := DC_PATH( cFileName )
IF EMPTY( cPath )
  cPath := SET(_SET_DEFAULT)
ELSE
  cPath := ''
ENDIF
IF VALTYPE(aArray)#'A'
  RETURN .F.
ENDIF
nHandle := FCREATE( cPath + cFileName )
IF nHandle<=0
  RETURN .f.
ENDIF
_DCARRAY_2 ( aArray, nHandle, '' )
FCLOSE(nHandle)
RETURN .T.

* -------------------

STATIC FUNCTION _dcarray_2 ( aArray , nHandle, cElement )

LOCAL  nElement, cTextLine, cType, nArrayLen, cValue

nArrayLen := LEN(aArray)
FWRITE(nHandle,cElement+',A,'+ALLTRIM(STR(nArrayLen))+CHR(13)+CHR(10))
FOR nElement := 1 TO nArrayLen
    IF VALTYPE(aArray[nElement])='U'
      LOOP
    ENDIF
    cValue := aArray[nElement]
    cType := VALTYPE(cValue)
    cTextLine := ''
    DO CASE
      CASE cType='C'
        cTextLine := STRTRAN(HARDCR(cValue),CHR(13)+CHR(10),CHR(255))
      CASE cType='N'
        cTextLine := ALLTRIM(STR(cValue))
      CASE cType='D'
        cTextLine := DTOC(cValue)
      CASE cType='L'
        cTextLine := IIF(cValue,'Y','N')
      CASE cType='A'
        _DCARRAY_2 ( aArray[nElement], nHandle, ;
                    cElement+'['+ALLTRIM(STR(nElement,4))+']')
        LOOP
    ENDCASE
    FWRITE(nHandle,cElement+'['+ALLTRIM(STR(nElement,4))+']';
           +','+cType+','+cTextLine+CHR(13)+CHR(10))
NEXT
RETURN .T.

* -------------------

* Convert a Ragged array to a Parallel array and vice-versa

FUNCTION dc_aconvert ( aInput )

LOCAL i, j, aOutput := {}

aOutput := Array( Len( aInput[1] ), Len( aInput ) )
FOR i := 1 TO LEN( aInput )
  FOR j := 1 TO LEN( aInput[1] )
    aOutput[j,i] := aInput[i,j]
  NEXT
NEXT
RETURN aOutput

* -------------------

* Same as ASCAN() except scan is in reverse order
* Code block is currently not used

FUNCTION dc_ascanrev ( aArray, xValue, nStart )

LOCAL i
nStart := iif(valtype(nStart)='N',nStart,LEN(aArray))
FOR i := nStart TO 1 STEP -1
  IF aArray[i] = xValue
    RETURN i
  ENDIF
NEXT
RETURN 0

* -------------------

* Reverses all elements in an array
FUNCTION dc_areverse ( aArray )

LOCAL nLength := LEN(aArray), i
LOCAL aOutArray
aOutArray := ARRAY(nLength)
FOR i := 1 TO nLength
  aOutArray[nLength-i+1] := aArray[i]
NEXT
RETURN aOutArray

* -------------------

FUNCTION dc_num2array ( nStart, nEnd, nIncr )

LOCAL aNum := ARRAY((nEnd-nStart+nIncr)/nIncr), i, nPointer := 1
FOR i := nStart TO nEnd STEP nIncr
  aNum[nPointer] := i
  nPointer++
NEXT
RETURN aNum

* -------------------

FUNCTION dc_abigelem ( aArray, nLength )

LOCAL i, nElement, nNewLength
nLength := 0
FOR i := 1 TO LEN( aArray )
  nNewLength := LEN( aArray[i] )
  IF nNewLength > nLength
    nElement := i
    nLength := nNewLength
  ENDIF
NEXT
RETURN nElement

* -------------------

FUNCTION dc_array2prg ( aArray, cFileName, cFunction, cComments, cCode )

LOCAL nHandle, cSaveScrn, nChoice := 1, aOutArray := {"",""},;
      nLevel := 2, nProcLength

cFileName := IIF(VALTYPE(cFileName)='C',UPPER(Alltrim(cFileName)),'')
IF !('.' $ cFileName )
  cFileName += '.PRG'
ENDIF
cFunction := IIF(Valtype(cFunction)='C',ALLTRIM(cFunction),cFileName)
IF EMPTY(cFunction)
  cFunction := cFileName
ENDIF
IF '.' $ cFunction
  cFunction := SubStr( cFunction, 1, AT('.',cFunction)-1 )
ENDIF
cFunction := StrTran(DC_Path( cFunction,.t. ),'()','')
IF VALTYPE(aArray)#'A'
  RETURN .F.
ENDIF
nHandle := FOPEN( cFileName, 2 )
IF nHandle <= 0
  nHandle := FCREATE( cFileName )
ELSE
  DC_MsgBox(,,{ cFileName, DC_ARRMSG_3_1,'' },,.f.,,.f.,@nChoice,;
      { DC_ARRMSG_3_2, DC_ARRMSG_3_3, DC_ARRMSG_3_4 } )
  IF nChoice = 0 .OR. nChoice = 3
    FCLOSE(nHandle)
    RETURN .f.
  ELSEIF nChoice = 2
    FCLOSE(nHandle)
    nHandle := FCREATE( cFileName )
  ENDIF
ENDIF
IF nHandle<=0
  RETURN .f.
ENDIF
FSEEK( nHandle, 0, 2 )
cSaveScrn := DC_WaitOn( {DC_ARRMSG_1,' '+cFileName,'',DC_ARRMSG_2} )
IF !EMPTY(cComments)
  aOutArray[1] += cComments
ENDIF
aOutArray[1] += 'FUNCTION ' + cFunction + CHR(13)+CHR(10)
aOutArray[1] += 'LOCAL a'+CHR(13)+CHR(10)
IF !EMPTY(cCode)
  aOutArray[1] += cCode
ENDIF
aOutArray[1] += ;
  '_' + RIGHT(cFunction,5)+ALLTRIM(STR(nLevel))+'(@a)'+CHR(13)+CHR(10)
aOutArray[nLevel] += ;
  'STATIC PROCEDURE _' + RIGHT(cFunction,5)+ALLTRIM(STR(nLevel))+'(a)'+;
  CHR(13)+CHR(10)
nProcLength := 0
_DCARRAY_W2 ( aArray, nHandle, '', aOutArray, @nLevel, @nProcLength, cFunction )
aOutArray[nLevel] += 'RETURN' + CHR(13)+CHR(10)
aOutArray[1] += 'RETURN a'+REPL(CHR(13)+CHR(10),2)
FOR nLevel := 1 TO LEN(aOutArray)
  FWRITE( nHandle, aOutArray[nLevel] )
NEXT
FCLOSE(nHandle)
DC_IMPL(cSaveScrn)
RETURN .T.

* -------------------

STATIC PROCEDURE _dcarray_w2 ( aArray , nHandle, cElement, aOutArray, ;
                               nLevel, nProcLength, cFunction )

LOCAL  nElement, cTextLine, cType, nArrayLen, cValue, i, n, nLength,;
       cDelim, j, cChar, cNewText, cNewLine, lOk, cSaveText

nArrayLen := LEN(aArray)
lOk := .t.
cSaveText := aOutArray[nLevel]
aOutArray[nLevel] += 'a' + cElement + ' := { ;' + CHR(13)+CHR(10)
FOR nElement := 1 TO nArrayLen
  cValue := aArray[nElement]
  cType := VALTYPE(cValue)
  cTextLine := ''
  DO CASE
    CASE LEN(aOutArray[nLevel]) > 2000
      lOk := .f.
      EXIT
    CASE cType$'UB'
      cTextLine := 'nil'
    CASE cType='C'
      cDelim := '"'
      IF LEN(cValue)=0
        cTextLine := cDelim + cDelim
      ELSE
        FOR i := 1 TO LEN(cValue) STEP 100
          cNewLine := SubStr(cValue,i,100)
          cNewText := ''
          FOR j := 1 TO LEN(cNewLine)
            cChar := SubStr(cNewLine,j,1)
            IF cChar>=CHR(0).AND.cChar<=CHR(31)
              cNewText += ;
                cDelim + '+CHR('+ALLTRIM(STR(ASC(cChar)))+')+'+ cDelim
            ELSEIF cChar='"'
              cNewText += cDelim + '+CHR(34)+'+ cDelim
            ELSE
              cNewText += cChar
            ENDIF
          NEXT
          cNewLine := cNewText
          cTextLine += cDelim + cNewLine
          IF i < LEN(cValue) .AND. LEN(cValue)>100
            cTextLine += cDelim + '+;'+CHR(13)+CHR(10)
          ELSE
            cTextLine += cDelim
          ENDIF
        NEXT
        IF RIGHT(cTextLine,1)=chr(10)
          cTextLine += '   ' + cDelim + SUBSTR(cValue,i,100) + cDelim
        ENDIF
      ENDIF
    CASE cType='N'
      cTextLine := ALLTRIM(STR(cValue))
    CASE cType='D'
      cTextLine := 'CTOD("'+DTOC(cValue)+'")'
    CASE cType='L'
      cTextLine := IIF(cValue,".T.",".F.")
    CASE cType='A'
      lOk := .f.
      EXIT
    OTHERWISE
      LOOP
  ENDCASE
  cTextLine := '  ' + ;
     cTextLine + IIF(nElement<nArrayLen,',','')+ ' ;' + CHR(13)+CHR(10)
  aOutArray[nLevel] += cTextLine
  nProcLength += LEN(cTextLine)


NEXT
IF lOk
  aOutArray[nLevel] += ' }' + CHR(13)+CHR(10)
  IF nProcLength > 10000
    aOutArray[nLevel] += 'RETURN' + REPL(CHR(13)+CHR(10),2)
    nLevel++
    nProcLength := 0
    aOutArray[1] += ;
      '_' + RIGHT(cFunction,5)+ALLTRIM(STR(nLevel))+'(@a)'+CHR(13)+CHR(10)
    AADD( aOutArray, 'STATIC PROCEDURE _' + RIGHT(cFunction,5) + ;
                     ALLTRIM(STR(nLevel))+'(a)'+ CHR(13)+CHR(10) )
  ELSE
    nLevel++
    AADD(aOutArray,'')
  ENDIF
  RETURN
ENDIF

aOutArray[nLevel] := cSaveText + 'a'+cElement+' := ARRAY('+;
                     ALLTRIM(STR(nArrayLen))+')'+CHR(13)+CHR(10)
FOR nElement := 1 TO nArrayLen
  IF VALTYPE(aArray[nElement])='U'
    LOOP
  ENDIF
  cValue := aArray[nElement]
  cType := VALTYPE(cValue)
  cTextLine := ''
  DO CASE
    CASE cType='C'
      cDelim := '"'
      IF LEN(cValue)=0
        cTextLine := cDelim + cDelim
      ELSE
        FOR i := 1 TO LEN(cValue) STEP 100
          cNewLine := SubStr(cValue,i,100)
          cNewText := ''
          FOR j := 1 TO LEN(cNewLine)
            cChar := SubStr(cNewLine,j,1)
            IF cChar>=CHR(0).AND.cChar<=CHR(31)
              cNewText += ;
                cDelim + '+CHR('+ALLTRIM(STR(ASC(cChar)))+')+'+ cDelim
            ELSEIF cChar='"'
              cNewText += cDelim + '+CHR(34)+'+ cDelim
            ELSE
              cNewText += cChar
            ENDIF
          NEXT
          cNewLine := cNewText
          cTextLine += cDelim + cNewLine
          IF i < LEN(cValue) .AND. LEN(cValue)>100
            cTextLine += cDelim + '+;'+CHR(13)+CHR(10)
          ELSE
            cTextLine += cDelim
          ENDIF
        NEXT
        IF RIGHT(cTextLine,1)=chr(10)
          cTextLine += '   ' + cDelim + SUBSTR(cValue,i,100) + cDelim
        ENDIF
      ENDIF
    CASE cType='N'
      cTextLine := ALLTRIM(STR(cValue))
    CASE cType='D'
      cTextLine := 'CTOD("'+DTOC(cValue)+'")'
    CASE cType='L'
      cTextLine := IIF(cValue,".T.",".F.")
    CASE cType='A'
      _DCARRAY_W2 ( aArray[nElement], nHandle, ;
           cElement+'['+ALLTRIM(STR(nElement,4))+']', aOutArray, ;
           @nLevel, @nProcLength, cFunction )
      LOOP
    OTHERWISE
      LOOP
  ENDCASE
  cTextLine := 'a'+cElement+'['+ALLTRIM(STR(nElement))+'] := ' ;
                       + cTextLine+CHR(13)+CHR(10)
  aOutArray[nLevel] += cTextLine
  nProcLength += LEN(cTextLine)
  IF nProcLength > 10000
    aOutArray[nLevel] += 'RETURN' + REPL(CHR(13)+CHR(10),2)
    nLevel++
    nProcLength := 0
    aOutArray[1] += ;
      '_' + RIGHT(cFunction,5)+ALLTRIM(STR(nLevel))+'(@a)'+CHR(13)+CHR(10)
    AADD( aOutArray, 'STATIC PROCEDURE _' + RIGHT(cFunction,5) + ;
                     ALLTRIM(STR(nLevel))+'(a)'+ CHR(13)+CHR(10) )

  ENDIF
NEXT
nLevel++
AADD(aOutArray,'')
RETURN

* -------------------

FUNCTION dc_acompare ( aArray1, aArray2, lDebug )

LOCAL nLength, i

DEFAULT lDebug := .f.

IF Valtype(aArray1)#'A' .OR. Valtype(aArray2)#'A'
  IF lDebug
    DCQQDEBUG  Valtype(aArray1), Valtype(aArray2)
  ENDIF
  RETURN .f.
ENDIF
IF LEN(aArray1) # LEN(aArray2)
  IF lDebug
    DCQQDEBUG Len(aArray1), Len(aArray2)
  ENDIF
  RETURN .f.
ENDIF
nLength := LEN(aArray1)

FOR i := 1 TO nLength

  IF Valtype(aArray1[i]) # Valtype(aArray2[i])
    IF lDebug
      DCQQDEBUG aArray1[i], aArray2[i]
    ENDIF
    RETURN .f.
  ELSEIF Valtype(aArray1[i]) = 'A'
    IF !DC_ACompare(aArray1[i],aArray2[i],lDebug)
      RETURN .f.
    ENDIF
  ELSEIF Valtype(aArray1[i]) = 'N' .and. Round(aArray1[i],10) # Round(aArray2[i],10)
    IF lDebug
      DCQQDEBUG aArray1[i], aArray2[i], i
    ENDIF
    RETURN .f.
  ELSEIF Valtype(aArray1[i]) # 'N' .and. !(aArray1[i]==aArray2[i])
    IF lDebug
      DCQQDEBUG aArray1[i], aArray2[i], i
    ENDIF
    RETURN .f.
  ENDIF

NEXT

RETURN .t.

* -------------------

FUNCTION dc_astrtran ( aArray, cString1, cString2 )

LOCAL i

IF Valtype(aArray)#'A'
  RETURN .f.
ENDIF
FOR i := 1 TO LEN(aArray)
  IF Valtype(aArray[i])='C'
    aArray[i] := Strtran(cString1,cString2)
  ELSEIF Valtype(aArray[i])='A'
    DC_Astrtran( aArray[i], cString1, cString2 )
  ENDIF
NEXT
RETURN .t.

* -------------------

FUNCTION DC_ArrayView( aArray, lExpand, lCollapse, lModal, cTitle, lSymmetrical )

LOCAL oTree1, GetList := {}, i, j, GetOptions, lWindows, oTree2, oSay2, ;
      lOk, oScrn, lReBuild := .f., lGui := DC_Gui(.t.), oDrawingArea1, ;
      oDrawingArea2, oDialog1, oDialog2, oStatic1, oStatic2, oDialog, ;
      lPrint := .f., oPrintButton, aSub, oBrowseButton, ;
      oMenu, oTree, oFindButton, nElement := 0, xFind := Space(30), ;
      cFind := 'C', lExactMatch := .f., lCaseSensitive := .f., ;
      oExpandButton

DEFAULT lExpand := .f., ;
        lCollapse := .f., ;
        lModal := .t., ;
        cTitle := 'Array / Object Browser', ;
        lSymmetrical := .f.

// PRIVATE __DCBDEBUG := .T.

IF 'DC_ARRAYVIEW' $ _CallStack() .AND. !('DC_INSPECTOBJECT' $ _CallStack())
  RETURN nil
ENDIF

IF Valtype(aArray) # 'A'
  aArray := { aArray }
  // DEFAULT lExpand := .t.
ENDIF

/*
BEGIN SEQUENCE
IF !lSymmetrical
  IF Len(aArray) > 0
    aSub := aArray[1]
    FOR i := 1 TO Len(aArray)
      IF !Valtype(aArray[i]) == 'A'
        BREAK
      ELSEIF !(Len(aArray[i]) == Len(aSub))
        BREAK
      ENDIF
      FOR j := 1 TO Len(aSub)
        IF !(Valtype(aSub[j])==Valtype(aArray[i,j]))
          BREAK
        ENDIF
      NEXT j
    NEXT i
    lSymmetrical := .t.
  ENDIF
ENDIF
END SEQUENCE
*/

IF lSymmetrical
  RETURN DC_ArrayEdit(aArray)
ENDIF

lWindows := .f.
DO WHILE .t.

  IF lExpand
    oScrn := DC_WaitOn()
  ENDIF

  lReBuild := .f.

  @ 0,40 DCSTATIC TYPE XBPSTATIC_TYPE_RECESSEDBOX SIZE 58,1 OBJECT oStatic2
  @ .1,1 DCSAY '' SAYSIZE 30,.8 PARENT oStatic2 OBJECT oSay2

  @ 1,40 DCDIALOG oDialog2 NOTITLEBAR SIZE 60,19 DRAWINGAREA oDrawingArea2 ;
        BORDER XBPDLG_DLGBORDER

  @ .1,.4 DCTREEROOT OBJECT oTree2 SIZE 58.6,18.5 FONT '10.Courier' ;
        HASLINES HASBUTTONS PARENT oDrawingArea2 ;
        ALWAYSSHOWSELECTION ;
        ITEMMARKED {|a,b,o|_ItemMarked(a,b,o,oTree2,oSay2)} ;
        EVAL {|o|o:keyboard := {|a,b,o|_TreeKeyboard(a,b,o)} }

  @ 0,1 DCSTATIC TYPE XBPSTATIC_TYPE_RECESSEDBOX SIZE 39,1 OBJECT oStatic1
  @ .1,1 DCSAY 'Entire Contents' SAYSIZE 30,.8 PARENT oStatic1

  @ 1,1 DCDIALOG oDialog1 NOTITLEBAR SIZE 40,19 DRAWINGAREA oDrawingArea1 ;
        EVAL {|o|o:resize:= ;
         {|a,b,o|_resizechild(a,b,oDialog1,oTree1,oDialog2,oStatic1,oStatic2, ;
                              oDialog,oTree2)} }

  @ .1,.4 DCTREEROOT OBJECT oTree1 SIZE 38.6,18.5 FONT '10.Courier' ;
        HASLINES HASBUTTONS PARENT oDrawingArea1 ;
        ALWAYSSHOWSELECTION ;
        ITEMMARKED {|a,b,o|_ItemMarked(a,b,o,oTree2,oSay2)} ;
        EVAL {|o|o:keyboard := {|a,b,o|_TreeKeyboard(a,b,o)} }

  @ 20.5,1 DCPUSHBUTTON CAPTION '&Expand' ;
        ACCELKEY {xbeK_ALT_E,Asc('E'),Asc('e')} ;
        OBJECT oExpandButton ;
        ACTION {||lReBuild := !lExpand, ;
                  IIF(lRebuild,DC_ReadGuiEvent(DCGUI_EXIT_OK,GetList),nil), ;
                  DC_PointerWait(), ;
                  DC_ExpandTree(oTree1), ;
                  DC_PointerArrow() } ;
        SIZE 10,1 ;
        TOOLTIP 'Expand Entire Tree'

  @ DCGUI_ROW, DCGUI_COL + 10 DCPUSHBUTTON CAPTION '&Collapse' ;
        ACCELKEY {xbeK_ALT_C,Asc('C'),Asc('c')} ;
        ACTION {||oTree1:lockUpdate(.t.), ;
                  DC_ExpandTree(oTree1,.f.), ;
                  oTree1:lockUpdate(.f.)} ;
        SIZE 10,1 ;
        TOOLTIP 'Collapse Entire Tree'

  @ DCGUI_ROW, DCGUI_COL + 10 DCPUSHBUTTON CAPTION '&Properties' SIZE 10,1 ;
            ACCELKEY {xbeK_ALT_P,Asc('P'),Asc('p')} ;
            ACTION {||_Properties(oTree1)}

  @ DCGUI_ROW, DCGUI_COL + 10 DCPUSHBUTTON CAPTION 'Prin~t' ;
        OBJECT oPrintButton ;
        ACCELKEY {xbeK_ALT_P,Asc('P'),Asc('p')} ;
        ACTION {||lReBuild := !lExpand, ;
                  IIF(lRebuild, ;
                      (lPrint:=.t.,DC_ReadGuiEvent(DCGUI_EXIT_OK,GetList)), ;
                      (DC_PrintTreeBranch(oTree1:rootItem),lPrint:=.f.)) } ;
        SIZE 10,1 ;
        TOOLTIP 'Print the Entire Expanded Tree'

  @ DCGUI_ROW,DCGUI_COL + 10 DCPUSHBUTTON CAPTION '~Browse' ;
        OBJECT oBrowseButton ;
        ACCELKEY {xbeK_ALT_B,Asc('B'),Asc('b')} ;
        ACTION {||DC_ArrayEdit(aArray)} ;
        SIZE 10,1 ;
        WHEN {||lSymmetrical} ;
        TOOLTIP 'This appears to be a symmetrical array, therefore browse it'

  @ DCGUI_ROW,DCGUI_COL + 10 DCPUSHBUTTON CAPTION '~Find' ;
        OBJECT oFindButton ;
        ACCELKEY {xbeK_ALT_F,Asc('F'),Asc('f')} ;
        ACTION {||_ArrayFind(oTree1, oTree1:rootItem,@nElement,@xFind,@cFind, ;
                             @lExactMatch, @lCaseSensitive,.t.)} ;
        SIZE 10,1 ;
        TOOLTIP 'Find a value in this array' ;
        WHEN {||lExpand}

  FOR i := 1 TO Len(aArray)
    _ArrayView( 1, GetList, aArray[i], @oTree1, @lWindows,, ;
                @lExpand, i, aArray )
  NEXT

  DCSUBMENU oMenu

  DCMENUITEM {||_MenuCaption(oTree)} ;
      ACTION {||_Properties(oTree)} ;
      PARENT oMenu

  DCGETOPTIONS HIDE

  DCREAD GUI EXPRESS FIT TO lOk ;
    OPTIONS GetOptions ;
    NOAUTORESTORE ;
    PARENT @oDialog ;
    _MODAL lModal ;
    _SETAPPWINDOW lModal ;
    TITLE cTitle ;
    EVAL {|o|SetAppWindow(o), ;
           o:resize := {|a,b,o|_ResizeWindow(a,b,o,oDialog1,oDialog2, ;
                                oStatic1,oStatic2,oTree1,oTree2)}, ;
           IIF( lExpand,DC_ExpandTree(oTree1,.t.),nil ), ;
           IIF( lCollapse,DC_ExpandTree(oTree1,.f.),nil ), ;
           IIF( Valtype(oScrn)='A',DC_Impl(oScrn),nil ), ;
           o:show(), ;
           IIF( lPrint,DC_PrintTreeBranch(oTree1:rootItem),nil), ;
           oTree1:rbDown := {|aPos,z,o|oTree := oTree1, oTree1:SetData(oTree1:itemFromPos(aPos)),DC_GetWhen(GetList), ;
                              oMenu:popup(o,aPos,2,XBPMENU_PU_DEFAULT+XBPMENU_PU_MOUSE_RBDOWN)},;
           oTree2:rbDown := {|aPos,z,o|oTree := oTree2, oTree2:SetData(oTree2:itemFromPos(aPos)),DC_GetWhen(GetList), ;
                              oMenu:popup(o,aPos,2,XBPMENU_PU_DEFAULT+XBPMENU_PU_MOUSE_RBDOWN)} }

  oDialog := nil
  oTree1 := nil
  oTree2 := nil
  oDialog1 := nil
  oDialog2 := nil
  oStatic1 := nil
  oStatic2 := nil

  IF !lReBuild
    EXIT
  ENDIF
  lExpand := .t.

ENDDO

DC_Gui(lGui)

RETURN nil

* --------------------

STATIC FUNCTION _ArrayFind( oTree, oTreeItem, nElement, xFind, cFind, ;
                            lExactMatch, lCaseSensitive, lPrompt )

LOCAL aItems, GetList[0], lTop := .t., i, lStatus := .t., nItem, xData, ;
      nFound := 0

IF lPrompt

  @ 0,0 DCSAY 'Value to find' GET xFind

  @ 2,0 DCRADIO cFind VALUE 'C' PROMPT '(C)haracter'
  @ 3,0 DCRADIO cFind VALUE 'N' PROMPT '(N)umeric'
  @ 4,0 DCRADIO cFind VALUE 'D' PROMPT '(D)ate'
  @ 5,0 DCRADIO cFind VALUE 'L' PROMPT '(L)ogical'

  @ 7,0 DCCHECKBOX lTop PROMPT 'Search from Top'
  @ 8,0 DCCHECKBOX lCaseSensitive PROMPT 'Search is Case Sensitive' ;
        WHEN {||cFind = 'C'}
  @ 9,0 DCCHECKBOX lExactMatch PROMPT 'Exact Match is Required' ;
        WHEN {||cFind = 'C'}

  DCREAD GUI FIT ADDBUTTONS TO lStatus TITLE 'Find an array element by a value'

  IF lTop
    nElement := 1
  ELSE
    nElement++
  ENDIF

  IF lStatus
    DC_ExpandTree( oTree )
  ENDIF

ENDIF

IF lStatus

  aItems := oTreeItem:getChildItems()

  FOR i := nElement TO Len(aItems)

    xData := DC_GetCargo(aItems[i])[1]
    nItem := DC_GetCargo(aItems[i])[4]

    IF Valtype(xData) == 'A'
      oTreeItem := aItems[i]
      nFound := _ArrayFind( oTree, oTreeItem, 1, xFind, cFind, ;
                            lExactMatch, lCaseSensitive, .f. )
      IF nFound > 0
        EXIT
      ELSE
        LOOP
      ENDIF

    ENDIF

    IF Valtype(xData) == cFind
      IF cFind == 'C'
        IF lCaseSensitive .AND. lExactMatch
          IF Trim(xFind) == xData
            nFound := nItem
            EXIT
          ENDIF
        ELSEIF !lCaseSensitive .AND. lExactMatch
          IF Upper(Trim(xFind)) == xData
            nFound := nItem
            EXIT
          ENDIF
        ELSEIF lCaseSensitive .AND. !lExactMatch
          IF xData = Trim(xFind)
            nFound := nItem
            EXIT
          ENDIF
        ELSEIF !lCaseSensitive .AND. !lExactMatch
          IF Upper(xData) = Upper(Trim(xFind))
            nFound := nItem
            EXIT
          ENDIF
        ENDIF
      ELSEIF cFind == 'D' .AND. CtoD(Trim(xFind)) == xData
        nFound := nItem
        EXIT
      ELSEIF cFind == 'N' .AND. Val(Trim(xFind)) == xData
        nFound := nItem
        EXIT
      ENDIF
    ENDIF
  NEXT

  IF nFound > 0
    oTree:setData(aItems[i])
  ENDIF

ENDIF

RETURN nFound

* --------------------

STATIC FUNCTION _MenuCaption( oTree )

LOCAL oItem, xValue, cCaption := '?????????????'

IF Valtype(oTree) == 'O'
  oItem := oTree:getData()
  xValue := oItem:cargo[3,1]
  IF Valtype(xValue) == 'O'
    cCaption := 'Inspect this Object'
  ELSEIF Valtype(xValue) == 'A'
    cCaption := 'Browse this Array in a new Window'
  ELSEIF Valtype(xValue) $ 'CM'
    cCaption := 'Edit the contents of this String'
  ELSEIF Valtype(xValue) == 'D'
    cCaption := 'Edit the contents of this Date'
  ELSEIF Valtype(xValue) == 'N'
    cCaption := 'Edit the contents of this Number'
  ELSEIF Valtype(xValue) == 'L'
    cCaption := 'Toggle this Logical Value'
  ENDIF
ENDIF

RETURN cCaption

* --------------------

STATIC FUNCTION _TreeKeyboard( mp1, mp2, oTreeRoot )

LOCAL oItem

IF mp1 == xbeK_ENTER

  oItem := oTreeRoot:getData()
  IF Valtype(oItem:cargo[3]) $ 'AO'
    IF oItem:cargo[3,3]
      IF oItem:isExpanded()
        oItem:expand(.f.)
      ELSE
        oItem:expand(.t.)
      ENDIF
    ELSE
      Eval(oItem:cargo[3,2],oItem)
    ENDIF
  ENDIF

ENDIF

RETURN nil

* --------------

STATIC FUNCTION _ResizeChild( aOldSize, aNewSize, oDialog1, oTree1, oDialog2, ;
                              oStatic1, oStatic2, oDialog, oTree2 )

oTree1:setSize({oTree1:currentSize()[1]+aNewSize[1]-aOldSize[1], ;
                oTree1:currentSize()[2]+aNewSize[2]-aOldSize[2]})

oStatic1:setSize({oDialog1:currentSize()[1]-5, ;
                  oStatic1:currentSize()[2]} )

oStatic1:setPos({oStatic1:currentPos()[1], ;
                 oDialog1:currentPos()[2]+oDialog1:currentSize()[2]} )

oDialog2:setPos({oDialog1:currentSize()[1],oDialog1:currentPos()[2]})

oDialog2:setSize({oDialog2:currentSize()[1]-(aNewSize[1]-aOldSize[1]), ;
                  oDialog1:currentSize()[2]})

oStatic2:setPos({oDialog2:currentPos()[1]+10, ;
                 oDialog2:currentPos()[2]+oDialog2:currentSize()[2]})

oStatic2:setsize({oDialog2:currentSize()[1]-10, ;
                  oStatic2:currentSize()[2]})

oTree2:setSize({oDialog2:currentSize()[1]-10, ;
                oDialog2:currentSize()[2]-10})

RETURN nil

* --------------

STATIC FUNCTION _ResizeWindow( aOldSize, aNewSize, oDialog, oDialog1, oDialog2, ;
                               oStatic1, oStatic2, oTree1, oTree2 )

oDialog2:setSize({oDialog2:currentSize()[1]+aNewSize[1]-aOldSize[1], ;
                  oDialog2:currentSize()[2]+aNewSize[2]-aOldSize[2]})

oTree2:setSize({oTree2:currentSize()[1]+aNewSize[1]-aOldSize[1], ;
                oTree2:currentSize()[2]+aNewSize[2]-aOldSize[2]})

oStatic2:setPos({oDialog2:currentPos()[1], ;
                 oDialog2:currentPos()[2]+oDialog2:currentSize()[2]})

oStatic2:setSize({oDialog2:currentSize()[1]-10, ;
                  oStatic2:currentSize()[2]} )

oStatic2:setPos({oDialog2:currentPos()[1]+10, ;
                 oDialog2:currentPos()[2]+oDialog2:currentSize()[2]} )

oDialog1:setSize({oDialog1:currentSize()[1], ;
                  oDialog1:currentSize()[2]+aNewSize[2]-aOldSize[2]})

oStatic1:setSize({oDialog1:currentSize()[1]-5, ;
                  oStatic1:currentSize()[2]} )

oStatic1:setPos({oStatic1:currentPos()[1], ;
                 oDialog1:currentPos()[2]+oDialog1:currentSize()[2]} )

RETURN nil

* ------------

STATIC FUNCTION _ItemMarked( oItem, aRect, oTree, oTree2, oSay2 )

LOCAL xCargo := oItem:cargo[3]
LOCAL bItemSelected

IF Valtype(xCargo) = 'A' .AND. Len(xCargo) >=6 .AND. Valtype(xCargo[6]) = 'B'
  Eval(xCargo[6],oItem,oTree,oTree2)
  oSay2:setCaption(oItem:caption)
ENDIF

RETURN nil

* --------------------

STATIC FUNCTION _ArrayView( nMode, GetList, xData, oTree, lWindows, ;
                            cCaption, lExpand, nElement, aArray )

LOCAL i, oThread, GetOptions, bAction, bExpand, cElement, oTreeItem, ;
      cString, cArray

DEFAULT nElement := 0
cElement := Alltrim(Str(nElement)) + ': '

IF nMode = 1

  IF Valtype(xData) $ 'CM'

     DCTREEITEM CAPTION cElement + Strtran(xData,'~','~~') PARENT oTree ;
       IMAGENORMAL ICON_CHARACTER ;
       IMAGEEXPANDED ICON_CHARACTER ;
       IMAGEMARKED ICON_CHARACTER ;
       CARGO { xData, nil, nil, nElement, aArray, {|a,b,c|_ArrayStats(a,b,c)} }

  ELSEIF Valtype(xData) = 'D'

     DCTREEITEM CAPTION cElement + DC_XtoC(xData) PARENT oTree ;
       IMAGENORMAL ICON_DATE ;
       IMAGEEXPANDED ICON_DATE ;
       IMAGEMARKED ICON_DATE ;
       CARGO { xData, nil, nil, nElement, aArray, {|a,b,c|_ArrayStats(a,b,c)} }

  ELSEIF Valtype(xData) = 'N'

     DCTREEITEM CAPTION cElement + DC_XtoC(xData) PARENT oTree ;
       IMAGENORMAL ICON_NUMERIC ;
       IMAGEEXPANDED ICON_NUMERIC ;
       IMAGEMARKED ICON_NUMERIC ;
       CARGO { xData, nil, nil, nElement, aArray, {|a,b,c|_ArrayStats(a,b,c)} }

  ELSEIF Valtype(xData) = 'L'

     DCTREEITEM CAPTION cElement + DC_XtoC(xData) PARENT oTree ;
       IMAGENORMAL ICON_LOGICAL ;
       IMAGEEXPANDED ICON_LOGICAL ;
       IMAGEMARKED ICON_LOGICAL ;
       CARGO { xData, nil, nil, nElement, aArray, {|a,b,c|_ArrayStats(a,b,c)} }

  ELSEIF Valtype(xData) = 'B'

     DCTREEITEM CAPTION cElement + DC_GetCodeBlockText(xData) PARENT oTree ;
       IMAGENORMAL ICON_CODEBLOCK ;
       IMAGEEXPANDED ICON_CODEBLOCK ;
       IMAGEMARKED ICON_CODEBLOCK ;
       CARGO { xData, nil, nil, nElement, aArray, {|a,b,c|_ArrayStats(a,b,c)} }

  ELSEIF Valtype(xData) = 'U'

     DCTREEITEM CAPTION cElement + 'Nil' PARENT oTree ;
       IMAGENORMAL ICON_NIL ;
       IMAGEEXPANDED ICON_NIL ;
       IMAGEMARKED ICON_NIL ;
       CARGO { xData, nil, nil, nElement, aArray, {|a,b,c|_ArrayStats(a,b,c)} }

  ELSEIF Valtype(xData) = 'O'

     cString := '   '
     IF IsMemberVar(xData,'CAPTION') .AND. !Empty(xData:caption)
       cString += '[' + DC_Xtoc(xData:caption) + '] '
     ENDIF
     IF IsMemberVar(xData,'TITLE') .AND. !Empty(xData:title)
       cString += '[' + DC_Xtoc(xData:title) + '] '
     ENDIF

     IF !xData:isDerivedFrom('XbpPartHandler') .AND. ;
        !xData:isDerivedFrom('XbpRoot') .AND. ;
        IsMemberVar(xData,'CARGO') .AND. ;
         Valtype(xData:cargo) == 'A' .AND. Len(xData:cargo) >= 3 .AND. ;
        Valtype(xData:cargo[2]) == 'A' .AND. Valtype(xData:cargo[1]) == 'N' .AND. !Empty(xData:cargo[1])
       IF !Empty(xData:cargo[2][xData:cargo[1],cGETLIST_TITLE])
         cString += '[TITLE:' + xData:cargo[2][xData:cargo[1],cGETLIST_TITLE] + '] '
       ENDIF
       IF !Empty(xData:cargo[2][xData:cargo[1],cGETLIST_ID])
         cString += '[ID:' + xData:cargo[2][xData:cargo[1],cGETLIST_ID] + '] '
       ENDIF
     ENDIF
     IF IsMethod(xData,'CURRENTPOS')
       cArray := DC_ArrayString(xData:currentPos())
       cArray := StrTran(cArray,'{')
       cArray := StrTran(cArray,'}')
       cString += '[P:'+cArray+'] '
     ENDIF
     IF IsMethod(xData,'CURRENTSIZE')
       cArray := DC_ArrayString(xData:currentSize())
       cArray := StrTran(cArray,'{')
       cArray := StrTran(cArray,'}')
       cString += '[S:'+cArray+'] '
     ENDIF

     cCaption := cElement + 'OBJECT:' + xData:ClassName() + cString
     bAction := { |o| _ArrayView( 2,  nil, o:cargo[3,1], o, @lWindows, ;
                                  cCaption, @lExpand, nElement, aArray ), ;
                      o:cargo[3,3] := .t., o:cargo[3,4] := nElement, ;
                      o:cargo[3,5] := aArray}

     DCTREEITEM CAPTION cCaption ;
       PARENT oTree ;
       CARGO { xData, bAction, .f., nElement, aArray, {|a,b,c|_ArrayStats(a,b,c)} } ;
       OBJECT oTreeItem ;
       ACTION bAction FIRSTCLICK ;
       IMAGENORMAL ICON_CLOSEDFOLDER ;
       IMAGEEXPANDED ICON_CLOSEDFOLDER ;
       IMAGEMARKED ICON_OPENFOLDER

     IF lExpand
       _ArrayView( 2, GetList, xData, @oTreeItem, @lWindows, ;
                   cCaption, @lExpand, nElement, aArray )
     ENDIF

  ELSEIF Valtype(xData) = 'A'

     cCaption := cElement + 'Array (' + Alltrim(Str(Len(xData))) + ')'
     bAction := { |o| _ArrayView( 2,  nil, o:cargo[3,1], o, @lWindows, ;
                                  cCaption, @lExpand, nElement, aArray ), ;
                      o:cargo[3,3] := .t., o:cargo[3,4] := nElement, ;
                      o:cargo[3,5] := aArray }

     DCTREEITEM CAPTION cCaption ;
       PARENT oTree ;
       CARGO { xData, bAction, .f., nElement, aArray, {|a,b,c|_ArrayStats(a,b,c)} } ;
       OBJECT oTreeItem ;
       ACTION bAction FIRSTCLICK ;
       IMAGENORMAL ICON_CLOSEDFOLDER ;
       IMAGEEXPANDED ICON_OPENFOLDER ;
       IMAGEMARKED ICON_OPENFOLDER

     IF lExpand
       _ArrayView( 2, GetList, xData, @oTreeItem, @lWindows, ;
                   cCaption, @lExpand, nElement, aArray )
     ENDIF

  ENDIF

ELSEIF nMode = 2 .AND. Valtype(xData) $ 'AO'

  IF Empty(GetList)
    GetList := {}
  ENDIF

  IF Valtype(xData) = 'O'
    IF IsMethod(xData,'CHILDLIST') .AND. xData:status() > 0
      xData := xData:ChildList()
    ELSE
      RETURN nil
    ENDIF
  ENDIF

  IF lWindows
    oTree := nil
    @ 1,1 DCTREEROOT OBJECT oTree SIZE 50,15 FONT '10.Courier' ;
      HASLINES HASBUTTONS
  ENDIF

  FOR i := 1 TO Len(xData)
    _ArrayView( 1, GetList, xData[i], @oTree, @lWindows,, ;
                @lExpand, i, aArray )
  NEXT

  IF !lExpand
    IF !lWindows
      // oTree:cargo[3,1] := nil
      DCREAD GUI EXPRESS EXIT PARENT oTree
      oTree:expand(.t.)
    ELSE
      oThread := Thread():new()
      Sleep(10)
      oThread:Start( {||_ArrayView(4,GetList,,@oTree,,cCaption)} )
    ENDIF
  ENDIF

ELSEIF nMode = 3

  oTree:setSize( { oTree:setParent():currentSize()[1]-30, ;
                   oTree:setParent():currentSize()[2]-90 } )

ELSEIF nMode = 4

   DCGETOPTIONS ;
     ORIGIN XBPDLG_ORIGIN_MOUSE ;
     NOBUSY ;
     HIDE

   DCREAD GUI EXPRESS FIT TITLE cCaption ;
     OPTIONS GetOptions ;
     EVAL {|o|o:resize := {||_ArrayView(5,,,@oTree)}, ;
              o:setPos({o:CurrentPos()[1],o:currentPos()[2]-o:currentSize()[2]} ), ;
              o:show() }

ELSEIF nMode = 5

  oTree:setSize( { oTree:setParent():currentSize()[1]-30, ;
                   oTree:setParent():currentSize()[2]-40 } )
ENDIF

RETURN nil

* --------------------

STATIC FUNCTION _ArrayStats( oItem, oTree, oTree2, oParent )

LOCAL GetList[0], aData := oItem:cargo[3,1], aArray := oItem:cargo[3,5], ;
      i, nElement := oItem:cargo[3,4], cElement, ;
      xData, cString, cCaption, bAction, oTreeItem, cArray, lWindows, ;
      lExpand

IF Valtype(aData) = 'O' .AND. IsMethod(aData,'ChildList') .AND. ;
   !(Upper(aData:className())=='XBPPARTHANDLER')
  aData := aData:childList()
ELSEIF Valtype(aData) = 'A'
ELSE
  oTree2:hide()
  oTree2:destroy()
  oTree2:create()
  oTree2:show()
  RETURN nil
ENDIF

FOR i := 1 TO Len(aData)

  xData := aData[i]
  cElement := Alltrim(Str(i)) + ': '

  IF Valtype(xData) $ 'CM'

     DCTREEITEM CAPTION cElement + xData ;
       IMAGENORMAL ICON_CHARACTER ;
       IMAGEEXPANDED ICON_CHARACTER ;
       IMAGEMARKED ICON_CHARACTER ;
       CARGO { xData, nil, nil, nElement, aArray }

  ELSEIF Valtype(xData) = 'D'

     DCTREEITEM CAPTION cElement + DC_XtoC(xData) ;
       IMAGENORMAL ICON_DATE ;
       IMAGEEXPANDED ICON_DATE ;
       IMAGEMARKED ICON_DATE ;
       CARGO { xData, nil, nil, nElement, aArray }

  ELSEIF Valtype(xData) = 'N'

     DCTREEITEM CAPTION cElement + DC_XtoC(xData) ;
       IMAGENORMAL ICON_NUMERIC ;
       IMAGEEXPANDED ICON_NUMERIC ;
       IMAGEMARKED ICON_NUMERIC ;
       CARGO { xData, nil, nil, nElement, aArray }

  ELSEIF Valtype(xData) = 'L'

     DCTREEITEM CAPTION cElement + DC_XtoC(xData) ;
       IMAGENORMAL ICON_LOGICAL ;
       IMAGEEXPANDED ICON_LOGICAL ;
       IMAGEMARKED ICON_LOGICAL ;
       CARGO { xData, nil, nil, nElement, aArray }

  ELSEIF Valtype(xData) = 'B'

     DCTREEITEM CAPTION cElement + DC_GetCodeBlockText(xData) ;
       IMAGENORMAL ICON_CODEBLOCK ;
       IMAGEEXPANDED ICON_CODEBLOCK ;
       IMAGEMARKED ICON_CODEBLOCK ;
       CARGO { xData, nil, nil, nElement, aArray }

  ELSEIF Valtype(xData) = 'U'

     DCTREEITEM CAPTION cElement + 'Nil' ;
       IMAGENORMAL ICON_NIL ;
       IMAGEEXPANDED ICON_NIL ;
       IMAGEMARKED ICON_NIL ;
       CARGO { xData, nil, nil, nElement, aArray }

  ELSEIF Valtype(xData) = 'O'

     cString := '   '
     IF IsMemberVar(xData,'CAPTION') .AND. !Empty(xData:caption)
       cString += '[' + DC_Xtoc(xData:caption) + '] '
     ENDIF
     IF IsMemberVar(xData,'TITLE') .AND. !Empty(xData:title)
       cString += '[' + DC_Xtoc(xData:title) + '] '
     ENDIF
     IF !xData:isDerivedFrom('XbpPartHandler') .AND. ;
        !xData:isDerivedFrom('XbpRoot') .AND. ;
        IsMemberVar( xData, 'CARGO' ) .AND. ;
        Valtype(xData:cargo) == 'A' .AND. Len(xData:cargo) >= 3 .AND. ;
        Valtype(xData:cargo[2]) == 'A' .AND. Valtype(xData:cargo[1]) == 'N' .AND. !Empty(xData:cargo[1])
       IF !Empty(xData:cargo[2][xData:cargo[1],cGETLIST_TITLE])
         cString += '[TITLE:' + xData:cargo[2][xData:cargo[1],cGETLIST_TITLE] + '] '
       ENDIF
       IF !Empty(xData:cargo[2][xData:cargo[1],cGETLIST_ID])
         cString += '[ID:' + xData:cargo[2][xData:cargo[1],cGETLIST_ID] + '] '
       ENDIF
     ENDIF
     IF IsMethod(xData,'CURRENTPOS')
       cArray := DC_ArrayString(xData:currentPos())
       cArray := StrTran(cArray,'{')
       cArray := StrTran(cArray,'}')
       cString += '[P:'+cArray+'] '
     ENDIF
     IF IsMethod(xData,'CURRENTSIZE')
       cArray := DC_ArrayString(xData:currentSize())
       cArray := StrTran(cArray,'{')
       cArray := StrTran(cArray,'}')
       cString += '[S:'+cArray+'] '
     ENDIF

     cCaption := cElement + 'OBJECT:' + xData:ClassName() + cString
     bAction := { |o| _ArrayStats( o, oTree, oTree2, o ) }

     DCTREEITEM CAPTION cCaption ;
       CARGO { xData, bAction, .f., nElement, aArray } ;
       ACTION bAction FIRSTCLICK ;
       IMAGENORMAL ICON_CLOSEDFOLDER ;
       IMAGEEXPANDED ICON_OPENFOLDER ;
       IMAGEMARKED ICON_OPENFOLDER

/*
     IF lExpand
       _ArrayView( 2, GetList, xData, @oTreeItem, @lWindows, ;
                   cCaption, @lExpand, nElement, aArray )
     ENDIF
*/

  ELSEIF Valtype(xData) = 'A'

     cCaption := cElement + 'Array (' + Alltrim(Str(Len(xData))) + ')'
     bAction := { |o| _ArrayStats( o, oTree, oTree2, o ) }

     DCTREEITEM CAPTION cCaption ;
       CARGO { xData, bAction, .f., nElement, aArray } ;
       ACTION bAction FIRSTCLICK ;
       IMAGENORMAL ICON_CLOSEDFOLDER ;
       IMAGEEXPANDED ICON_OPENFOLDER ;
       IMAGEMARKED ICON_OPENFOLDER

/*
     IF lExpand
       _ArrayView( 2, GetList, xData, @oTreeItem, @lWindows, ;
                   cCaption, @lExpand, nElement, aArray )
     ENDIF
*/

  ENDIF

NEXT

IF oParent == nil
  oTree2:hide()
  oTree2:destroy()
  oTree2:create()
  oTree:show()
  oParent := oTree2
ENDIF

oTree2:lockUpdate(.t.)
DCREAD GUI EXPRESS PARENT oParent EXIT
oTree2:lockUpdate(.f.)

IF oParent:isDerivedFrom('XbpTreeViewItem')
  oParent:expand(.t.)
ENDIF

RETURN nil

* ----------------------

STATIC FUNCTION _Properties(oTannenBaum)

LOCAL oItem := oTannenBaum:getData(), cValue, nDecimals, lOk, ;
      nLength, xValue, cVarName, nElement, aArray, GetList := {}, ;
      cElement, oThread

IF Valtype(oItem) == 'O'
  oTannenBaum:setData(oItem)
  xValue := oItem:cargo[3,1]
  nElement := oItem:cargo[3,4]
  aArray := oItem:cargo[3,5]
  cElement := Alltrim(Str(nElement)) + ': '
  lOk := .f.
  IF Valtype(xValue) = 'O'
    DC_InspectObject(xValue)
  ELSEIF Valtype(xValue) = 'A'
    oThread := Thread():new()
    Sleep(5)
    oThread:start({||DC_ArrayView(xValue)})
  ELSEIF Valtype(xValue) = 'N'
    cValue := Str(xValue)
    IF '.' $ cValue
      nDecimals := Len(cValue) - At('.',cValue)
    ELSE
      nDecimals := 0
    ENDIF
    @ 1,1 DCSAY 'Numeric Value' GET xValue PICT '999999999999.999999999999' SAYSIZE 0
    @ 3,1 DCSAY 'Decimals' GET nDecimals PICT '99' SAYSIZE 0
    DCREAD GUI EXPRESS FIT ADDBUTTONS TO lOk ENTEREXIT TITLE 'Edit Numeric Value' MODAL
    IF lOk
      xValue := Val(Str(xValue,,nDecimals))
    ENDIF
  ELSEIF Valtype(xValue) = 'D'
    @ 1,1 DCSAY 'Date Value' GET xValue POPUP {|d|DC_PopDate(d,,,,cVarName,,2)} ;
      SAYSIZE 0
    DCREAD GUI EXPRESS FIT ADDBUTTONS TO lOk ENTEREXIT TITLE 'Edit Date Value' MODAL
  ELSEIF Valtype(xValue) = 'C'
    nLength := Len(xValue)
    @ 1,1 DCSAY 'Character Value' SAYSIZE 0 FONT '12.Courier Bold'
    @ 2,1 DCMULTILINE xValue SIZE 80,10 FONT '10.Courier'
    @ 13,1 DCSAY 'String Length (0=Trim)' GET nLength PICT '9999' SAYSIZE 0
    DCREAD GUI EXPRESS FIT ADDBUTTONS TO lOk ENTEREXIT TITLE 'Edit Character Value' MODAL
    IF lOk
      IF nLength = 0
        xValue := Trim(xValue)
      ELSE
        xValue := Pad(xValue,nLength)
      ENDIF
    ENDIF
  ELSEIF Valtype(xValue) = 'L'
    @ 1,1 DCCHECKBOX xValue PROMPT 'Logical Value'
    DCREAD GUI EXPRESS FIT ADDBUTTONS TITLE 'Edit Logical Value' MODAL TO lOk
  ENDIF
  IF lOk
    aArray[nElement] := xValue
    oItem:cargo[3,1] := xValue
    oItem:setCaption( cElement + DC_XtoC(xValue) )
  ENDIF

ENDIF

RETURN nil

* ----------------------

FUNCTION DC_ExpandTree( oTree, lExpand )

LOCAL i, aChildList, oTannenBaum := oTree

DEFAULT lExpand := .t.

IF IsMethod(oTannenBaum,'LOCKUPDATE')
  oTannenBaum:lockUpdate(.t.)
ENDIF

IF oTree:isDerivedfrom('XbpTreeView')
   oTree := oTree:rootItem
ENDIF

aChildList := oTree:getChildItems()

FOR i := 1 TO Len(aChildList)
  aChildList[i]:expand(lExpand)
  DC_ExpandTree( aChildList[i], lExpand )
NEXT

IF IsMethod(oTannenBaum,'LOCKUPDATE')
  oTannenBaum:lockUpdate(.f.)
ENDIF

RETURN nil

* -----------------------

GETSETFUNCTION DC_InspectObjectEvent DEFAULT 0

* -----------------------

FUNCTION DC_InspectObject( oXbp, lDisplayResults, aResults, lModal )

LOCAL aIvars, aSlots, i, j, cString, GetList := {}, GetOptions, ;
      aMethods, aIvarResults := {}, aSlotResults := {}, ;
      aMethodResults := {}, bErrorBlock, oError, cInfo, ;
      cResults, aInfo, cVarInfo, oIvarBrowse, oSlotBrowse, ;
      oMethodBrowse, aBrowPres, nIvarPointer, nSlotPointer, ;
      nMethodPointer, xValue

PRIVATE oObject := oXbp
DEFAULT lDisplayResults := .t., ;
        lModal := .t.

aResults := oXbp:classDescribe()

aIvars := {}
FOR i := 1 TO Len(aResults[3])
  AAdd(aIvars,Alltrim(aResults[3,i,1]))
NEXT
ASort(aIvars,,,{|x,y|Upper(x)<Upper(y)})
aMethods := {}
FOR i := 1 TO Len(aResults[4])
  AAdd(aMethods,Alltrim(aResults[4,i,1])+'()')
NEXT
ASort(aMethods,,,{|x,y|Upper(x)<Upper(y)})
aSlots := {}

FOR j := 1 TO 3

  IF j = 1
    cString := 'Instance Variables:' + Chr(13)+Chr(10)+Chr(13)+Chr(10)
    aInfo := aIvars
    aResults := aIvarResults
  ELSEIF j = 2
    cString +=  Chr(13) + Chr(10) + 'Callback Slots:' + Chr(13)+Chr(10)+Chr(13)+Chr(10)
    aInfo := aSlots
    aResults := aSlotResults
  ELSEIF j = 3
    cString +=  Chr(13) + Chr(10) + 'Methods:' + Chr(13)+Chr(10)+Chr(13)+Chr(10)
    aInfo := aMethods
    aResults := aMethodResults
  ENDIF

  FOR i := 1 TO Len(aInfo)

    cVarInfo := _VarInfo(oXbp, aInfo[i])

    IF !Empty(cVarInfo)
      xValue := nil
      cResults := ''
      bErrorBlock := ERRORBLOCK( {|oError| Break(oError)} )
      IF j < 3
        BEGIN SEQUENCE
          xValue := &('oObject:' + aInfo[i])
          cResults := DC_XtoC( xValue, .t., 200 )
        RECOVER USING oError
          cResults := Alltrim(DC_XtoC(oError:genCode)) + ": " + ;
                      oError:description
        END SEQUENCE
        ERRORBLOCK(bErrorBlock)
      ENDIF
      cString += Pad(aInfo[i],20) + ' = ' + DC_XtoC(cResults) + ' Info: ' + ;
                   DC_XtoC(cVarInfo) + Chr(13) + Chr(10)
      AAdd(aResults, { aInfo[i], cResults, cVarInfo, xValue } )
    ENDIF

  NEXT

NEXT j

cString := Strtran(cString,Chr(0),'')

aResults := { aIvarResults, aSlotResults, aMethodResults }

IF lDisplayResults

  aBrowPres := ;
  { { XBP_PP_COL_HA_FGCLR, GRA_CLR_WHITE },            /* Header FG Color  */  ;
    { XBP_PP_COL_HA_BGCLR, GRA_CLR_DARKGRAY },         /* Header BG Color  */  ;
    { XBP_PP_COL_DA_ROWSEPARATOR, XBPCOL_SEP_DOTTED }, /* Row Sep          */  ;
    { XBP_PP_COL_DA_COLSEPARATOR, XBPCOL_SEP_DOTTED }, /* Col Sep          */  ;
    { XBP_PP_COL_DA_FGCLR, GRA_CLR_BLACK },            /* Row FG Color     */  ;
    { XBP_PP_COL_DA_BGCLR, GRA_CLR_WHITE },            /* Row BG Color     */  ;
    { XBP_PP_COL_DA_ROWHEIGHT, 11 },                   /* Row Height       */  ;
    { XBP_PP_HILITE_FGCLR, GRA_CLR_WHITE },            /* Hilite FG color  */  ;
    { XBP_PP_HILITE_BGCLR, GRA_CLR_DARKBLUE },         /* Hilite BG color  */  ;
    { XBP_PP_COL_FA_FGCLR, GRA_CLR_WHITE }             /* Footer FG Color  */  ;
  }

  @ 1,0 DCBROWSE oIvarBrowse DATA aIvarResults ;
        SIZE 80,IIF(Empty(aSlots),12,7) ;
        PRESENTATION aBrowPres NOSOFTTRACK FONT '8.Alaska Crt' ;
        POINTER nIvarPointer ;
        ITEMSELECTED {||_DisplayObject(aIvarResults[nIvarPointer,4])}
  DCBROWSECOL ELEMENT 1 HEADER 'Variable Name' WIDTH 20 PARENT oIvarBrowse
  DCBROWSECOL ELEMENT 2 HEADER 'Value' WIDTH 30 PARENT oIvarBrowse
  DCBROWSECOL ELEMENT 3 HEADER 'Info' WIDTH 50 PARENT oIvarBrowse

  IF !Empty(aSlots)
    @ 8,0 DCBROWSE oSlotBrowse DATA aSlotResults SIZE 80,7 ;
          PRESENTATION aBrowPres NOSOFTTRACK FONT '8.Alaska Crt' ;
          POINTER nSlotPointer ;
          ITEMSELECTED {||_DisplayObject(aSlotResults[nSlotPointer,4])}
    DCBROWSECOL ELEMENT 1 HEADER 'Slot Name' WIDTH 20 PARENT oSlotBrowse
    DCBROWSECOL ELEMENT 2 HEADER 'Value' WIDTH 30 PARENT oSlotBrowse
    DCBROWSECOL ELEMENT 3 HEADER 'Info' WIDTH 50 PARENT oSlotBrowse
  ENDIF

  @IIF(Empty(aSlots),13,15),0 DCBROWSE oMethodBrowse DATA aMethodResults ;
        SIZE 80,IIF(Empty(aSlots),12,7) ;
        PRESENTATION aBrowPres NOSOFTTRACK FONT '8.Alaska Crt' ;
        POINTER nMethodPointer ;
        ITEMSELECTED {||_CallMethod(aMethodResults[nMethodPointer,1],oXbp)}
  DCBROWSECOL ELEMENT 1 HEADER 'Method Name' WIDTH 20 PARENT oMethodBrowse
  DCBROWSECOL ELEMENT 2 HEADER 'Value' WIDTH 30 PARENT oMethodBrowse
  DCBROWSECOL ELEMENT 3 HEADER 'Info' WIDTH 50 PARENT oMethodBrowse

  DCGETOPTIONS AUTORESIZE

  DCREAD GUI EXPRESS ;
     FIT ;
     _MODAL lModal ;
     OPTIONS GetOptions ;
     _SETAPPWINDOW lModal ;
     TITLE 'Object Inspector: ' + oXbp:Classname() ;
     EVAL {|o|SetAppWindow(o)}

ENDIF

RETURN cString

* ----------------

STATIC FUNCTION _CallMethod( cMethod, oXbp )

LOCAL xData, oThread
PRIVATE _oXbp := oXbp

xData := &('_oXbp:'+cMethod)

IF Valtype(xData) == 'A'
  oThread := Thread():new()
  Sleep(10)
  oThread:Start({||DC_ArrayView(xData,,,.f.)})
ELSEIF Valtype(xData) == 'O'
  IF IsMethod(xData,'CHILDLIST')
    oThread := Thread():new()
    Sleep(10)
    oThread:Start({||DC_ArrayView(xData,,,.f.)})
  ELSE
    oThread := Thread():new()
    Sleep(10)
    oThread:Start({||DC_InspectObject(xData,,,.f.)})
  ENDIF
ELSEIF Valtype(xData) $ 'CM'
  DC_GuiMemoEdit(xData)
ELSEIF Valtype(xData) == 'D'
  DC_PopDate(xData)
ELSEIF Valtype(xData) $ 'NFI'
  DC_PopCalc(xData)
ELSEIF Valtype(xData) = 'L'
  DCMSGBOX IIF(xData,'Yes','No')
ENDIF

RETURN nil

* ----------------

STATIC FUNCTION _DisplayObject( oObject )

IF Valtype( oObject ) = 'A'
  DC_ArrayView( oObject )
ELSEIF Valtype( oObject ) = 'O'
  DC_InspectObject( oObject)
ENDIF

RETURN nil

* ----------------

STATIC FUNCTION _VarInfo(oXbp, iVar)

LOCAL cRet := ""
iVar := Strtran(iVar,'()','')

IF IsMemberVar(oXbp, iVar, VAR_ASSIGN_PROTECTED)
  cRet += "VAR_ASSIGN_PROTECTED "
ENDIF
IF IsMemberVar(oXbp, iVar, VAR_ASSIGN_EXPORTED)
  cRet += "VAR_ASSIGN_EXPORTED "
ENDIF
IF IsMemberVar(oXbp, iVar, VAR_ASSIGN_HIDDEN)
  cRet += "VAR_ASSIGN_HIDDEN "
ENDIF
IF IsMemberVar(oXbp, iVar, VAR_INSTANCE)
  cRet += "VAR_ASSIGN_INSTANCE "
ENDIF
IF IsMemberVar(oXbp, iVar, VAR_CLASS)
  cRet += "VAR_CLASS "
ENDIF
IF IsMemberVar(oXbp, iVar, VAR_CLASS_SHARED)
  cRet += "VAR_CLASS_SHARED "
ENDIF
IF IsMemberVar(oXbp, iVar, CLASS_HIDDEN)
  cRet += "VAR CLASS_HIDDEN "
ENDIF
IF IsMemberVar(oXbp, iVar, CLASS_PROTECTED)
  cRet += "VAR CLASS_PROTECTED "
ENDIF
IF IsMemberVar(oXbp, iVar, CLASS_EXPORTED)
  cRet += "VAR CLASS_EXPORTED "
ENDIF

IF IsMethod(oXbp, iVar, METHOD_INSTANCE)
  cRet += "METHOD_INSTANCE "
ENDIF
IF IsMethod(oXbp, iVar, METHOD_CLASS)
  cRet += "METHOD_CLASS "
ENDIF
IF IsMethod(oXbp, iVar, CLASS_EXPORTED)
  cRet += "METHOD CLASS_EXPORTED "
ENDIF
IF IsMethod(oXbp, iVar, CLASS_HIDDEN)
  cRet += "METHOD CLASS_HIDDEN "
ENDIF
IF IsMethod(oXbp, iVar, CLASS_PROTECTED)
  cRet += "METHOD CLASS_PROTECTED "
ENDIF

cRet := StrTran(cRet,"VAR_ASSIGN_EXPORTED VAR_ASSIGN_INSTANCE VAR CLASS_EXPORTED", "(standard ivar)")
cRet := StrTran(cRet,"METHOD_INSTANCE METHOD CLASS_EXPORTED","(standard method)")
cRet := StrTran(cRet,"(standard ivar) (standard method)","(standard callback/assign)")

RETURN cRet

* -------------------

STATIC FUNCTION _PrintTree( oTree )

DC_WinAlert('The Tree Print function is not completed. Sorry!')

RETURN nil

* ------------------

FUNCTION DC_ArraySet( aArray, nElement, xValue )

aArray[nElement] := xValue

RETURN nil

* -------------------

FUNCTION DC_PrintTreeBranch( oTreeItem, nIndent, cFont, lPreview )

LOCAL oPrinter

DEFAULT lPreview := .t., ;
        nIndent := 3, ;
        cFont := '10.Courier New'

DCPRINT ON ;
  _PREVIEW lPreview ;
  TO oPrinter ;
  NONSTOP

IF Valtype(oPrinter) # 'O' .OR. !oPrinter:lActive
  RETURN nil
ENDIF

_PrintTreeBranch( oTreeItem, oPrinter, 5, nIndent )

DCPRINT OFF ;
  PRINTER oPrinter

RETURN nil

* -------------------

STATIC FUNCTION _PrintTreeBranch( oTreeItem, oPrinter, nCol, nIndent )

LOCAL i, aTreeItems := oTreeItem:getChildItems()

@ DC_PrinterRow(oPrinter)+1, nCol ;
   DCPRINT SAY oTreeItem:caption ;
   PRINTER oPrinter

IF DC_PrinterRow(oPrinter) > 55
  DCPRINT EJECT PRINTER oPrinter
ENDIF

FOR i := 1 TO Len(aTreeItems)
  _PrintTreeBranch( aTreeItems[i], oPrinter, nCol + nIndent, nIndent )
NEXT

RETURN nil

* ------------------

STATIC FUNCTION _CallStack()

LOCAL aStack := {}, i := 2

DO WHILE .t.
  AAdd( aStack, Upper(ProcName(i)) )
  i++
  IF Empty(ProcName(i))
    EXIT
  ENDIF
ENDDO

RETURN aStack

* -----------------

FUNCTION DC_ArrayEdit( aArray )

LOCAL GetList[0], GetOptions, oBrowse, aSub, i, cPicture, bSort

@ 0,0 DCSAY 'Double-Click headings to optimize columns' ;
      SAYSIZE 0 RESIZE DCGUI_RESIZE_REPOSONLY_Y

@ 1,0 DCBROWSE oBrowse DATA aArray SIZE 80,19 ;
      PRESENTATION DC_BrowPres() ;
      EDIT xbeBRW_ItemSelected ;
      INSERT xbeK_INS ;
      DELETE xbeK_DEL ;
      RESIZE DCGUI_RESIZE_RESIZEONLY ;
      FONT '9.Lucida Console' ;
      OPTIMIZE

aSub := aArray[1]
FOR i := 1 TO Len(aSub)
  IF Valtype(aSub[i]) == 'N'
    cPicture := '9999999999.9999'
  ELSE
    cPicture := ''
  ENDIF
  bSort := SortBlock(aArray,i)
  DCBROWSECOL ELEMENT i HEADER Alltrim(Str(i)) ;
     WIDTH Len(DC_XtoC(aSub[i])) PARENT oBrowse ;
     PICTURE cPicture SORT bSort
NEXT

@ 20.5,0 DCPUSHBUTTON CAPTION 'Add New (INS)' SIZE 13,1.2 ;
   ACTION {||PostAppEvent(xbeP_Keyboard,xbeK_INS,,oBrowse)}

@ DCGUI_ROW, DCGUI_COL + 10 DCPUSHBUTTON ;
   CAPTION 'Delete (DEL)' SIZE 13,1.2 ;
   ACTION {||PostAppEvent(xbeP_Keyboard,xbeK_DEL,,oBrowse)}

@ DCGUI_ROW, DCGUI_COL + 10 DCPUSHBUTTON ;
   CAPTION 'Exit' SIZE 9, 1.2 ;
   ACTION {||DC_ReadGuiEvent(DCGUI_EXIT_OK,GetList)}

DCGETOPTIONS RESIZE

DCREAD GUI ;
   FIT ;
   OPTIONS GetOptions ;
   TITLE 'Browsing Array' ;
   MODAL ;
   SETAPPWINDOW

RETURN aArray

* ------------

STATIC FUNCTION SortBlock( aArray, i )

RETURN {||ASort(aArray,,,{|a,b|a[i]<b[i]})}
